import warnings
from collections.abc import Callable, Collection
from functools import partial
from numbers import Integral, Real
from typing import Any, Optional

import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure, SubFigure
from matplotlib.scale import ScaleBase, scale_factory
from scipy.optimize import curve_fit, least_squares
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.exceptions import DataConversionWarning
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import (
    _check_sample_weight,  # type: ignore
    check_array,
    check_is_fitted,
    check_non_negative,
    column_or_1d,
)

from ._backend import gmean, np
from ._plot import get_fig_ax, subplots
from .core import (
    _get_value_or_key,
    coerce_array,
    deprecate,
    dropna,
    flatten,
    geothmetic_meandian,
    match_kwargs,
    std,
)
from .model import MODELS, Model


@coerce_array
def _slice_dim(
    a,
    axis: Optional[int | Collection[int]] = None,
    keep: int = 0,
):
    """Slice an array along one or more dimensions.

    Parameters
    ----------
    a : ArrayLike
        The input array to be sliced.
    axis : None or int or tuple of ints, optional
        The dimension(s) along which to slice the array. If None, all
        dimensions except the first will be sliced. If an integer, a
        single dimension will be sliced. If a collection of integers,
        multiple dimensions will be sliced. Default is None.
    keep : int, default 0
        The index to keep along the sliced dimensions. Default is 0.

    Returns
    -------
    ArrayLike
        The sliced array.
    """
    if a.ndim == 0:
        return a
    if a.ndim == 1:
        return a[keep]
    if axis is None:
        axis = tuple(range(1, a.ndim))
    elif not isinstance(axis, Collection):
        axis = (axis,)
    indexer = tuple(keep if dim in axis else slice(None) for dim in range(a.ndim))
    return a[indexer]


_AGGREGATION_STRATEGIES = {
    "median": np.median,
    "mean": np.average,
    "average": np.average,
    "geomean": gmean,
    "gmean": gmean,
    "min": np.min,
    "max": np.max,
    "geothmetic_meandian": geothmetic_meandian,
    "gmnd": geothmetic_meandian,
    "first": partial(_slice_dim, keep=0),
    "last": partial(_slice_dim, keep=-1),
}

_WEIGHTING_SCHEMES = {
    **{key: lambda X, y: np.ones_like(y) for key in ["1", "1.0", "none", "None"]},
    "1/y": lambda X, y: 1 / y,
    "1/y^2": lambda X, y: 1 / y**2,
    "1/X": lambda X, y: 1 / X,
    "1/X^2": lambda X, y: 1 / X**2,
}


class CalCurve(BaseEstimator, RegressorMixin, TransformerMixin):
    """Calibration Curve transformer and regressor.

    Parameters
    ----------
    model : Model or str, default="4PL"
        The model to use for the calibration curve. Can be an instance
        of a Model or a string representing the model name. Current
        available options are:
        - "linear" : Linear function.
        - "power" : Power function.
        - "Hill" : Hill function.
        - "logistic" : Logistic function.
        - "3PL" : Four-parameter logistic (3PL) function.
        - "4PL" : Four-parameter logistic (4PL) function.
        - "5PL" : Five-parameter logistic (5PL) function.

    agg_reps : str Callable, or None, default="median"
        Aggregation method for replicates. Can be a string representing
        an aggregation strategy or a callable function. Current
        available options are:
        - "median"
        - "mean"
        - "average"
        - "geomean", "gmean"
        - "min"
        - "max"
        - "geothmetic_meandian", "gmnd"
        - "first"
        - "last"
        If None, then no aggregation is performed.

    coef_init : array-like, optional
        Initial coefficients for the model.

    warm_start : bool, default=False
        Whether to reuse the solution of the previous call to fit as
        initialization.

    solver : str, default="trf"
        Solver to use for optimization. Options are "trf", "dogbox", or
        "lm".

    lod_sds : float, default=3
        Number of standard deviations for limit of detection
        calculation.

    max_iter : int, optional
        Maximum number of iterations for the solver.

    ensure_2d : bool, default=False
        Whether to ensure the input is 2-dimensional.

    Attributes
    ----------
    coef_ : ndarray
        Coefficients of the fitted model.

    n_iter_ : int
        Number of iterations run by the solver.

    lod_ : float
        Calculated limit of detection.
    """

    _parameter_constraints = {
        # see docs for _param_validation.validate_parameter_constraints
        "model": [Model, StrOptions(set(MODELS))],
        "agg_reps": [callable, StrOptions(set(_AGGREGATION_STRATEGIES))],
        "coef_init": ["array-like", None],
        "warm_start": ["boolean"],
        "solver": [StrOptions({"trf", "dogbox", "lm"})],
        "lod_sds": [Interval(Real, 0, None, closed="left")],
        "max_iter": [Interval(Integral, 0, None, closed="left"), None],
        "ensure_2d": ["boolean"],
        "sample_weight": [StrOptions(set(_WEIGHTING_SCHEMES)), None],
    }

    def __init__(
        self,
        model: Model | str = "4PL",
        agg_reps: Optional[str | Callable] = "median",
        coef_init: Optional[np.ndarray] = None,
        warm_start: bool = False,
        solver: str = "trf",
        lod_sds: float = 3,
        max_iter: Optional[int] = None,
        ensure_2d: bool = False,
        sample_weight: Optional[str] = "1/y",
        **kwargs,
    ):
        self.model = model
        self.agg_reps = agg_reps
        self.coef_init = coef_init
        self.warm_start = warm_start
        self.solver = solver
        self.lod_sds = lod_sds
        self.max_iter = max_iter
        self.ensure_2d = ensure_2d
        self.sample_weight = sample_weight

        if kwargs:
            warnings.warn(
                f"The keyword arguments {kwargs.keys()} are not recognized. "
                + "Version 1.0 introduced breaking changes to the CalCurve class. "
                + "Please refer to the documentation for the most recent API.",
                DeprecationWarning,
            )

    def _get_sample_weight(self, X, y, sample_weight=None):
        if sample_weight is None:
            sample_weight = self.sample_weight
        if isinstance(sample_weight, str) and sample_weight in _WEIGHTING_SCHEMES:
            sample_weight = _WEIGHTING_SCHEMES[sample_weight](X, y)
        return column_or_1d(sample_weight, warn=True)

    def _aggregate_replicates(
        self, X: np.ndarray, y: np.ndarray, sample_weight: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        blank = np.ravel(y[X == 0])

        if self.agg_reps is None:
            return X, y, sample_weight, blank

        unique_X = np.unique(X, axis=0)
        agg_reps = _get_value_or_key(_AGGREGATION_STRATEGIES, self.agg_reps)
        unique_weight = []
        for x in unique_X:
            sw_x = sample_weight[x == X]
            unique_weight.append(agg_reps(sw_x) * len(sw_x))
        unique_weight = np.array(unique_weight)
        unique_X = unique_X[unique_weight > 0]
        unique_weight = unique_weight[unique_weight > 0]
        unique_y = np.array([agg_reps(y[x == X]) for x in unique_X])

        return unique_X, unique_y, unique_weight, blank

    def _preprocess_fit_data(self, X, y, sample_weight=None):
        # check X, y
        X = self._validate_data(
            X=X,
            force_all_finite=False,
            ensure_2d=self.ensure_2d,
            ensure_min_samples=len(self._model.coef_init),
        )
        y = check_array(
            y, force_all_finite=False, ensure_2d=False, estimator=self, input_name="y"
        )
        if np.ndim(X) > 1:
            if self.n_features_in_ > 1 and not self.ensure_2d:  # type: ignore
                warnings.warn(
                    "CalCurve input X should be a 1d array or 2d array with 1 feature; "
                    f"using only the first of {self.n_features_in_} features.",  # type: ignore
                    DataConversionWarning,
                )
            X = X[:, 0]

        # get sample weight
        sample_weight = self._get_sample_weight(X, y, sample_weight)

        # drop NANs and aggregate replicates
        X, y, sample_weight = dropna(X, y, sample_weight)
        self.X_, self.y_ = self._validate_data(X, y, ensure_2d=False)
        if self.ensure_2d:
            check_non_negative(self.X_, whom=self.__class__.__name__)
        sample_weight = _check_sample_weight(sample_weight, self.X_)
        return self._aggregate_replicates(self.X_, self.y_, sample_weight)

    def fit(self, X, y, sample_weight=None):
        """Fit the model to data.

        Parameters
        ----------
        X : array-like of shape (n_samples,) or (n_samples, n_features)
            Training data features, e.g., concentrations.

        y : array-like of shape (n_samples,)
            Target signal values, e.g., AEB or fluorescence intensity.

        sample_weight : str or array-like, optional
            Sample weights.

        Returns
        -------
        CalCurve
            Fitted estimator.
        """
        # parameters
        self._validate_params()  # type: ignore
        self._model = _get_value_or_key(MODELS, self.model)

        # data
        X, y, sample_weight, blank = self._preprocess_fit_data(X, y, sample_weight)

        # coef_init
        if self.warm_start and hasattr(self, "coef_"):
            coef_init = self.coef_
        elif self.coef_init is None:
            coef_init = self._model.coef_init
        else:
            coef_init = self.coef_init

        # optimization
        def loss_func(coefs, *, X, sample_weight):
            predictions = self._model.func(X, *coefs)
            residuals = predictions - y
            return residuals * sample_weight

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            result = least_squares(
                loss_func,
                x0=coef_init,
                jac=self._model.jac,
                max_nfev=self.max_iter,
                method=self.solver,
                kwargs={"X": X, "sample_weight": sample_weight},
            )
        self.coef_ = result.x
        self.n_iter_ = result.nfev

        self.lod_ = limit_of_detection(blank, self, lod_sds=self.lod_sds)
        return self

    def signal(self, X):
        """Predict the signal (e.g., AEB) for given concentrations.

        Parameters
        ----------
        X : array-like
            Input data of concentrations.

        Returns
        -------
        signal : array-like
            Predicted signal values.
        """
        return self._model.func(X, *self.coef_)

    def conc(self, y):
        """Estimate the concentration for given signal values.

        Parameters
        ----------
        y : array-like
            Signal values.

        Returns
        -------
        conc : array-like
            Estimated concentration values.
        """
        return self._model.inverse(y, *self.coef_)

    def predict(self, X):
        """Predict signal using the calibration curve model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or (n_samples,)
            Input data of concentrations.

        Returns
        -------
        y : ndarray
            Predicted signal values.
        """
        check_is_fitted(self)
        X = self._validate_data(
            X=X, reset=False, force_all_finite=False, ensure_2d=self.ensure_2d
        )
        if np.ndim(X) == 1:
            return self.signal(X)
        else:
            return self.signal(X[:, 0])

    def transform(self, X):
        """Transform concentrations into signal.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or (n_samples,)
            Input data of concentrations.

        Returns
        -------
        y : ndarray
            Transformed signal values.
        """
        check_is_fitted(self)
        X = self._validate_data(
            X=X, reset=False, force_all_finite=False, ensure_2d=self.ensure_2d
        )
        # if self.ensure_2d:
        #     X = X[:, 0].reshape(-1, 1)
        return self.signal(X)

    def inverse_transform(self, X):
        """Estimate the concentration for given signal values.

        Parameters
        ----------
        y : array-like of shape (n_samples, n_features) or (n_samples,)
            Signal values.

        Returns
        -------
        conc : array-like
            Back-calculated concentration values.
        """
        # (n_samples, n_features) -> (n_samples, n_features)
        check_is_fitted(self)
        X = self._validate_data(
            X=X, reset=False, force_all_finite=False, ensure_2d=self.ensure_2d
        )
        return self.conc(X)

    def _make_curve_points(self, **kwargs):
        """Make the points to interpolate for plotting a calibration curve.

        Parameters
        ----------
        cc : CalCurve
            Fitted calibration curve.

        Returns
        -------
        X_curve, y_curve
            Points for plotting the calibration curve.

        Other Parameters
        ----------------
        xscale : str or matplotlib.scale.ScaleBase, optional
            Override the calibration curve's x-scale.
        num : int, default 50
            Number of samples to generate. Default is 50. Must be
            non-negative.

        """
        check_is_fitted(self)
        xscale = kwargs.get("xscale", self._model.xscale)
        if not isinstance(xscale, ScaleBase):
            xscale = scale_factory(
                scale=xscale,
                axis=None,  # type: ignore
                **match_kwargs(scale_factory, kwargs),  # type: ignore
            )

        trans = xscale.get_transform()
        vmin = np.amin(self.X_)
        vmax = np.amax(self.X_)
        minpos = np.amin(self.X_[self.X_ > 0])
        if np.isfinite(self.lod_):
            vmin = min(vmin, self.lod_)
            minpos = min(minpos, self.lod_)
        vmin, vmax = xscale.limit_range_for_scale(vmin=vmin, vmax=vmax, minpos=minpos)
        trvmin = trans.transform(vmin)
        trvmax = trans.transform(vmax)
        start = 1.1 * trvmin - 0.1 * trvmax
        stop = 1.1 * trvmax - 0.1 * trvmin
        X_curve = trans.inverted().transform(
            np.linspace(start=start, stop=stop, num=kwargs.get("num", 50))
        )
        y_curve = self.signal(X_curve)
        self._curve_points = (X_curve, y_curve)
        calculated = {
            "xscale": xscale,
            "trans": trans,
            "vmin": vmin,
            "vmax": vmax,
            "trvmin": trvmin,
            "trvmax": trvmax,
            "start": start,
            "stop": stop,
            "X_curve": X_curve,
            "y_curve": y_curve,
        }
        return calculated

    def plot(
        self,
        fig: Optional[Figure | SubFigure] = None,
        ax: Optional[Axes] = None,
        **kwargs,
    ):
        """Plot the calibration curve and calibrator points.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure object. If None, the current figure will be
            used. If there is no current figure, a new figure will be
            created.
        ax : matplotlib.axes.Axes, optional
            The axes object. If None, the current axes will be used. If
            there is no current axes, a new axes will be created.
        **kwargs
            Additional keyword arguments for customizing the plot, such
            as `xlabel`, `ylabel`, and `title`.

        """
        check_is_fitted(self)
        self._make_curve_points(**kwargs)

        fig, ax = get_fig_ax(fig, ax)
        X_curve, y_curve = self._curve_points

        lod_kwargs = {
            "x": self.lod_,
            "color": "tab:blue",
            "linestyle": "--",
            "alpha": 0.5,
            "label": f"limit of detection: {self.lod_:.2g}",
        } | match_kwargs("lod_", kwargs)
        if np.isfinite(lod_kwargs["x"]):
            ax.axvline(lod_kwargs.pop("x"), **lod_kwargs)

        curve_kwargs = {
            "x": X_curve,
            "y": y_curve,
            "color": "tab:green",
            "label": f"{self._model.name} calibration curve",
        } | match_kwargs("curve_", kwargs)
        ax.plot(curve_kwargs.pop("x"), curve_kwargs.pop("y"), **curve_kwargs)

        point_kwargs = {
            "x": self.X_,
            "y": self.y_,
            "color": "black",
            "linestyle": "",
            "marker": "o",
            "alpha": 0.5,
            "label": "calibrator points",
        } | match_kwargs("point_", kwargs)
        ax.plot(point_kwargs.pop("x"), point_kwargs.pop("y"), **point_kwargs)

        ax_kwargs = (
            {
                "xscale": self._model.xscale,
                "yscale": self._model.yscale,
            }
            | match_kwargs(ax.set, kwargs)
            | match_kwargs("ax_", kwargs)
        )
        ax.set(**ax_kwargs)

        legend_kwargs = match_kwargs("legend_", kwargs)
        ax.legend(**legend_kwargs)

    def _more_tags(self):
        if self.ensure_2d:
            return {"allow_nan": True, "poor_score": True, "requires_positive_X": True}
        else:
            return {
                "allow_nan": True,
                "requires_positive_X": True,
                "X_types": ["2darray", "1darray"],
            }

    # Deprecated methods
    @deprecate()
    def bound_lod(self, x_flat):
        return np.maximum(x_flat, self.lod_)

    @deprecate("CalCurve.signal, CalCurve.predict, or CalCurve.transform")
    def fun(self, x):
        try:
            y = self._model.func(self.bound_lod(x), **self.coef_)
        except TypeError:
            x_flat = np.array(x)
            y = self._model.func(self.bound_lod(x_flat), **self.coef_)
        return y

    @deprecate("CalCurve.conc or CalCurve.inverse_transform")
    def inverse(self, y):
        try:
            x = self.bound_lod(self._model.inverse(y, **self.coef_))
        except TypeError:
            y_flat = np.array(y)
            x = self.bound_lod(self._model.inverse(y_flat, **self.coef_))
        return x

    @deprecate("CalCurve.fit")
    @classmethod
    def from_data(
        cls,
        *,
        x,
        y,
        model,
        lod_sds=3,
        corr="c4",
        force_lod: bool = False,
        weights="1/y^2",
        **kwargs,
    ):
        cc = cls(model=model, lod_sds=lod_sds)
        return cc.fit(x, y, sample_weight=weights)

    @deprecate()
    @classmethod
    def from_function(
        cls,
        fun,
        inverse,
        lod=-np.inf,
        lod_sds=3,
        force_lod: bool = False,
        xscale="linear",
        yscale="linear",
    ):
        pass


@deprecate("CalCurve.fit")
def regress(*, x, y, model, weights="1/y^2", **kwargs):
    named_model = model if isinstance(model, Model) else MODELS[model]
    sigma = None
    if isinstance(weights, str):
        if weights == "1/y^2":
            sigma = y
        if weights == "1/y":
            sigma = np.sqrt(np.array(y))
        elif weights == "1":
            sigma = np.ones(len(y))
    if sigma is None:
        sigma = flatten(weights) ** -2
    calibration_function = named_model.func
    xdata, ydata, sigma = dropna([x, y, sigma])
    return curve_fit(
        f=calibration_function, xdata=xdata, ydata=ydata, sigma=sigma, **kwargs
    )[0]


def limit_of_detection(
    blank: Any,
    inverse: CalCurve | Model | Callable | str,
    lod_sds: float = 3,
    corr: str = "c4",
    coef=None,
    nan_policy="omit",
    **kwargs,
):
    """Computes the limit of detection (LOD).

    Parameters
    ----------
    blank : array-like
        Signal (e.g., average number of enzymes per bead, AEB) of the
        zero calibrator. Must have at least two elements.
    inverse_fun : ``function`` or ``CalCurve``
        The functional form used for the calibration curve. If a
        function, it should accept the measurement reading (`y`, e.g.,
        fluorescence) as its only argument and return the value (`x`,
        e.g., concentration). If **inverse_fun** is a ``CalCurve``
        object, the LOD will be calculated from its ``inverse`` method.
    sds : numeric, optional
        How many standard deviations above the mean should the
        background should the limit of detection be calculated at?
        Common values include 2.5 (Quanterix), 3 (Walt Lab), and 10
        (lower limit of quantification, LLOQ).
    corr : {"n", "n-1", "n-1.5", "c4"} or numeric, default "c4"
        The sample standard deviation under-estimates the population
        standard deviation for a normally distributed variable.
        Specifies how this should be addressed. Options:

            - "n" : Divide by the number of samples to yield the
              uncorrected sample standard deviation.

            - "n-1" : Divide by the number of samples minus one to
              yield the square root of the unbiased sample variance.

            - "n-1.5" : Divide by the number of samples minus 1.5 to
              yield the approximate unbiased sample standard deviation.

            - "c4" : Divide by the correction factor to yield the
              exact unbiased sample standard deviation.

            - If numeric, gives the delta degrees of freedom.

    Returns
    -------
    numeric
        The limit of detection, in units of x (e.g., concentration).

    See Also
    --------
    c4 : factor `c4` for unbiased estimation of the standard deviation

    std : unbiased estimate of the population standard deviation

    numpy.std : standard deviation

    """
    blank = np.ravel(np.asarray(blank))

    if nan_policy == "omit":
        blank = blank[~np.isnan(blank)]

    if len(blank) < 2:  # noqa: PLR2004
        # warnings.warn(f"limit_of_detection undefined for {len(blank)} replicates")
        return np.nan

    lod_signal = lod_sds * std(blank, corr=corr) + np.mean(blank)

    if isinstance(inverse, CalCurve):
        f = inverse.conc
    elif isinstance(inverse, Model):
        f = inverse.inverse
    elif callable(inverse):
        f = inverse
    else:
        f = _get_value_or_key(MODELS, inverse, (CalCurve, Model, Callable)).inverse

    if coef is None:
        return f(lod_signal, **kwargs)  # kwargs can be provided or not
    elif not kwargs:
        return f(lod_signal, *coef)  # only if coef is provided
    else:
        raise ValueError("Coefficients can be given in coef or kwargs, but not both.")


class _CalCurveSeries(pd.Series):
    """Subclass of pandas.Series for when entries are CalCurves."""

    @property
    def _constructor(self):
        return _CalCurveSeries

    def __getitem__(self, key):
        result = super().__getitem__(key)
        return self.__class__(result) if isinstance(result, pd.Series) else result

    # def subplots(
    #     self, nrows: int = 1, ncols: int = 1, **kwargs
    # ) -> tuple[Figure, np.ndarray]:
    #     """Create a figure and axes grid.

    #     Parameters
    #     ----------
    #     nrows, ncols : int, default 1
    #         Number of rows/columns of the subplot grid.

    #     **kwargs
    #         Additional keyword arguments passed to `plt.subplots()`.

    #     Returns
    #     -------
    #     fig : matplotlib.figure.Figure
    #         The created figure object.
    #     axs : matplotlib.axes.Axes or array of Axes
    #         ax can be either a single Axes object, or an array of Axes
    #         objects if more than one subplot was created.

    #     """
    #     figsize = (
    #         ncols * rcParams["figure.figsize"][0],
    #         nrows * rcParams["figure.figsize"][1],
    #     )
    #     return plt.subplots(
    #         nrows=nrows,
    #         ncols=ncols,
    #         figsize=figsize,
    #         squeeze=False,
    #         **match_kwargs(plt.subplots, kwargs),
    #     )

    def plot(  # type: ignore
        self,
        *args,
        **kwargs,
    ) -> Axes | np.ndarray:
        """Plot the calibration curve and calibrator points.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure object. If None, the current figure will be
            used. If there is no current figure, a new figure will be
            created.
        ax : matplotlib.axes.Axes, optional
            The axes object. If None, the current axes will be used. If
            there is no current axes, a new axes will be created.
        **kwargs
            Additional keyword arguments for customizing the plot, such
            as `xlabel`, `ylabel`, and `title`.

        """
        fig = kwargs.get("fig")
        ax = kwargs.get("ax")
        max_cols = kwargs.get("max_cols", 5)

        levels = [
            i
            for i in range(self.index.nlevels)
            if self.index.get_level_values(i).nunique(dropna=False) > 1
        ]

        if len(levels) > 1:
            cols = self.index.get_level_values(levels[0]).unique()
            ncols = self.index.get_level_values(levels[0]).nunique()
            nrows = self.index.get_level_values(0).value_counts().max()
            fig, axs = subplots(nrows=nrows, ncols=ncols, fig=fig, **kwargs)

            ax_s = pd.Series(index=self.index)
            for c, col in enumerate(cols):
                # idx = sorted(self.index[self.index.get_level_values(levels[0]) == col])
                # ax_s[col] = dict(zip(idx, axs[:, c]))
                ax_s[col] = axs[: len(ax_s[col]), c]
        else:
            nrows = int(np.ceil(len(self) / max_cols))
            ncols = int(np.ceil(len(self) / nrows))
            fig, axs = subplots(nrows=nrows, ncols=ncols, fig=fig, **kwargs)
            ax_s = pd.Series(dict(zip(self.index, np.ravel(axs))))

        for i, ax in ax_s.items():
            title = "\n".join(str(a) for a in i) if isinstance(i, tuple) else i
            self[i].plot(ax=ax, title=title, **match_kwargs(self[i].plot, kwargs))

        return ax_s.item() if len(ax_s) == 1 else axs  # type: ignore
        # self_notna = self[self.notna()]

        # if len(self_notna) == 0:
        #     # empty series
        #     ax_map = {}

        # elif len(self_notna) == 1:
        #     # only one CalCurve in series
        #     if ax is not None and isinstance(ax, Axes):
        #         if fig is None:
        #             fig = ax.figure
        #     elif fig is not None:
        #         ax = fig.subplots(**match_kwargs(plt.subplots, kwargs))
        #     else:
        #         fig, ax = self.subplots(**kwargs)
        #     ax_map = {self.index[0]: ax}

        # elif self.index.nlevels == 1:
        #     # only one assay or only one plex
        #     if len(self) <= max_cols:
        #         fig, axs = self.subplots(ncols=len(self_notna), **kwargs)
        #         ax_map = {self_notna.index[i]: axs[i] for i in range(len(self_notna))}
        #     else:
        #         nrows = int(np.ceil(len(self) / max_cols))
        #         fig, axs = self.subplots(nrows=nrows, ncols=max_cols, **kwargs)
        #         ax_map = {
        #             self_notna.index[i]: axs[divmod(i, max_cols)]
        #             for i in range(len(self_notna))
        #         }

        # else:
        #     # try to lay out cal curves according to assay and plex
        #     ncols = self.index.get_level_values(0).nunique()
        #     nrows = self.index.get_level_values(0).value_counts().max()
        #     fig, axs = self.subplots(nrows=nrows, ncols=ncols, **kwargs)
        #     ax_map = {}
        #     for a0, assay_level_0 in enumerate(
        #         sorted(self_notna.index.get_level_values(0).unique())
        #     ):
        #         for a1, assay_level_1 in enumerate(
        #             sorted(self_notna[assay_level_0].index)
        #         ):
        #             ax_map[(assay_level_0, assay_level_1)] = axs[a1, a0]

        # for assay, ax in ax_map.items():
        #     if hasattr(assay, "__iter__") and not isinstance(assay, str):
        #         title = "\n".join([str(a) for a in assay])
        #     else:
        #         title = assay
        #     self[assay].plot(
        #         ax=ax, title=title, **match_kwargs(self[assay].plot, kwargs)
        #     )  # type: ignore
        # return

        # # if tight_layout:
        # #     plt.tight_layout()



import sys
from collections.abc import Callable

_MODULES_CHECK: dict[str, bool] = {
    "jax": "jax" in sys.modules,
    "numba": True,
}

if _MODULES_CHECK["jax"]:
    import jax.numpy as np
    from jax import jit
    from jax.scipy.special import gammaln

    def gmean(a, axis: int = 0, dtype=None, weights=None):
        """Geometric mean."""
        return np.exp(np.average(np.log(a), axis=axis, weights=weights))

    import warnings

    warnings.warn("JAX backend is experimental and may not work as expected.")  # type: ignore
else:
    import numpy as np
    from scipy.special import gammaln
    from scipy.stats import gmean

    try:
        from numba import jit  # type: ignore
    except ImportError:
        _MODULES_CHECK["numba"] = False

        def jit(fun: Callable) -> Callable:
            """Dummy jit decorator for when numba/jax not installed."""
            return fun



"""Functions for analyzing MOSAIC data.

In addition to the dependencies for waltlabtools, waltlabtools.mosaic
also requires pandas 0.25 or greater and scikit-learn 0.21 or greater.

The public functions in waltlabtools.mosaic can be accessed via, e.g.,

.. code-block:: python

   import waltlabtools as wlt  # waltlabtools main functionality
   import waltlabtools.mosaic  # for analyzing MOSAIC assays

   subset_data = wlt.mosaic.plate_subsets()  # analyze data from a plate

if also using other functionality from the waltlabtools package, or

.. code-block:: python

   from waltlabtools import mosaic  # for analyzing MOSAIC assays

   subset_data = mosaic.plate_subsets()  # analyze data from a plate

if using only the waltlabtools.mosaic module.


-----


"""

import os
import re
import warnings
from tkinter import filedialog

import pandas as pd
from numpy.random import default_rng
from sklearn.mixture import GaussianMixture

from ._backend import jit, np
from .cal_curve import CalCurve, limit_of_detection
from .core import _DDOFS, c4, flatten

__all__ = [
    "PlateFileCollection",
    "mixture_orientation",
    "weighted_aeb",
    "log_transform",
    "mixture_aeb",
    "well_to_aeb",
    "extended_coefs",
    "plate_subsets",
]


# CONSTANTS
TOL = 1e-4
REG_COVAR = 1e-6
N_INIT = 5
MAX_ITER = 100
BACKGROUND_LEVEL = 2000
coefs_index = ["a", "b", "c", "d", "LOD", "LLOQ", "blank mean", "blank std", "blank cv"]


# CLASSES
class PlateFileCollection:
    """Collection of RCA product fluorescence intensity files.

    A PlateFileCollection object is a container for MOSAIC files. It is
    used to keep all wells from a given day, calibration curve, or assay
    together.

    Parameters
    ----------
    dir_path : str, optional
        The directory containing the MOSAIC plate files. If not provided
        a file dialog will be opened to select the folder.

    Attributes
    ----------
    name : str
        The name of the folder containing the MOSAIC plate files.
    wells : list
        The wells contained in the MOSAIC plate files, e.g., "A1".
    file_map : dict
        A dictionary mapping the well position to the file path for
        the corresponding flow cytometry output file.
    conc_map : dict
        A dictionary mapping the well position to the known
        concentration of the calibrator.
    dir_path
        The path to the folder containing the MOSAIC plate files.

    """

    def __init__(self, dir_path=None):
        if dir_path is None:
            dir_path_ = filedialog.askdirectory(
                title="Choose a Folder With Data for One Calibration Curve"
            )
        else:
            dir_path_ = dir_path

        layout_file = None
        well_files = []
        for file_entry in os.scandir(dir_path_):
            if file_entry.name.endswith(".xlsx") and ("layout" in file_entry.name):
                layout_file = file_entry.path
            elif file_entry.name.endswith(".csv"):
                well_files.append(file_entry)
        if layout_file is None:
            raise OSError("Could not find layout file in folder " + dir_path_ + ".")

        name = re.split(r"[/\\\\]", dir_path_)[-1].replace(" FINAL", "")

        file_map = {}
        conc_map = (
            pd.read_excel(layout_file, header=None, index_col=0).squeeze().to_dict()
        )
        for file_entry in well_files:
            split_filename = re.split(r"[-–—_ .()\[\]|,*]", file_entry.name)
            for well in conc_map:
                if well in split_filename:
                    if file_entry.path not in file_map.values():
                        file_map[well] = file_entry.path
                    else:
                        file_map[file_entry.name] = file_entry.path

        if conc_map.keys() != file_map.keys():
            missing_files = set(conc_map.keys()) - set(file_map.keys())
            if missing_files:
                warnings.warn(
                    "In the folder '"
                    + name
                    + "', for the following wells, no CSV data files were found: '"
                    + "', '".join(missing_files)
                    + "'."
                )
                for key in missing_files:
                    del conc_map[key]

            missing_concs = set(file_map.keys()) - set(conc_map.keys())
            if missing_concs:
                warnings.warn(
                    "In the folder '"
                    + name
                    + "', the following wells were not found in the layout file: '"
                    + "', '".join(missing_concs)
                    + "'."
                )
                for key in missing_concs:
                    del file_map[key]

        self.name = name
        self.wells = sorted(conc_map.keys())
        self.file_map = file_map
        self.conc_map = conc_map
        self.dir_path = dir_path_


@jit
def mixture_orientation(
    means_: np.ndarray, covariances_: np.ndarray, threshold_sds=5
) -> tuple:
    """Determines which peak is 'off.'

    Parameters
    ----------
    means_ : ndarray of length 2
        The two means of the Gaussian mixture model.
    covariances_ : ndarray of length 2
        The two covariances of the Gaussian mixture model.
    threshold_sds : float, default 5
        The number of standard deviations above the mean to use as a
        threshold for distinguishing on-beads in the case where there
        are very few on-beads.

    Returns
    -------
    off_label, sds_threshold, ordered_means
        Tuple of length 3. Its elements:

        - off_label : 1 or 2
            The label of the Gaussian mixture model with the lower mean.

        - sds_threshold : float
            The number of standard deviations above the mean to use as a
            threshold for distinguishing on-beads in the case where
            there are very few on-beads.

        - ordered_means : ndarray of length 2
            The two means of the Gaussian mixture model, ordered from
            lowest to highest (off to on).

    """

    off_label = np.argmin(means_)
    off_on = np.array([off_label, 1 - off_label])
    low_sd = np.sqrt(covariances_[off_label])
    sds_threshold = means_[off_label] + threshold_sds * low_sd
    ordered_means = means_[off_on]
    return off_label, sds_threshold, ordered_means


@jit
def weighted_aeb(onfrac_gmm, onfrac_sds):
    """Weighted average of the two measures of fraction on.

    Calculates the AEB using a weighted average of the Gaussian
    mixture model fraction on and the standard deviation-based fraction
    on.

    Parameters
    ----------
    onfrac_gmm : float
        The on-fraction of the Gaussian mixture model.
    onfrac_sds : float
        The on-fraction based on the standard deviation threshold.

    Returns
    -------
    aeb : float
        The average number of enzymes per bead.

    See Also
    --------
    aeb : calculate the AEB from a single f_on value

    """
    gmm_weight = (0.5 - 0.5 * np.cos(np.pi * onfrac_gmm)) ** 2
    onfrac = (1 - gmm_weight) * onfrac_sds + gmm_weight * onfrac_gmm
    return -np.log(1 - onfrac)


def log_transform(flat_data: np.ndarray) -> np.ndarray:
    """Log-transforms the data.

    Because a few values may be negative, a constant value is calculated
    to add to all of the values.

    Parameters
    ----------
    flat_data : 1D ndarray
        The data to be transformed.

    Returns
    -------
    log_data : 1D ndarray
        The log-transformed data.

    """
    first_percentile = np.quantile(flat_data, 0.01)
    const = first_percentile - BACKGROUND_LEVEL
    return np.log(flat_data - const)


def mixture_aeb(
    flat_data: np.ndarray,
    means_init=None,
    flat_len=None,
    threshold_sds=5,
    reg_covar=REG_COVAR,
) -> float:
    """Calculates AEB based on a 2-Gaussian mixture model.

    Parameters
    ----------
    flat_data : 1D ndarray
        Data used for fitting the mixture model. It is assumed that if
        the data should be log-transformed, they have already been
        transformed, e.g., with `log_transform`.
    means_init : array-like of length 2, optional
        The user-provided initial means, If not provided, means are
        initialized as the maximum and minimum of the data.
    flat_len : int, optional
        Length of flat_data. If not provided, it is calculated.
    threshold_sds : numeric, default 5
        The number of standard deviations above the mean to use as a
        threshold for distinguishing on-beads in the case where there
        are very few on-beads.
    reg_covar : numeric, default REG_COVAR
        Non-negative regularization added to the diagonal of covariance.
        Allows to assure that the covariance matrices are all positive.

    Returns
    -------
    float
        AEB, based on weighted average of two measures of calculating
        on-fraction.

    """
    mi = [np.amin(flat_data), np.amax(flat_data)] if means_init is None else means_init
    means_init_ = np.array(mi).reshape(-1, 1)
    mixture = GaussianMixture(
        n_components=2,
        means_init=means_init_,
        tol=TOL,
        reg_covar=reg_covar,
        n_init=N_INIT,
        max_iter=MAX_ITER,
        covariance_type="spherical",
    )
    try:
        gmm_labels = mixture.fit_predict(flat_data.reshape(-1, 1))
        off_label, sds_threshold, ordered_means = mixture_orientation(
            np.ravel(mixture.means_), np.ravel(mixture.covariances_), threshold_sds
        )
        flat_labels = np.ravel(1 - gmm_labels) if off_label else np.ravel(gmm_labels)
        corrected_flat_labels = (flat_labels & (flat_data > ordered_means[0])) | (
            flat_data > ordered_means[1]
        )
        flat_len_ = flat_len if flat_len is not None else len(flat_data)
        onfrac_gmm = np.sum(corrected_flat_labels) / flat_len_
        onfrac_sds = np.sum(flat_data > sds_threshold) / flat_len_
        return weighted_aeb(onfrac_gmm, onfrac_sds)
    except Exception as exc:
        if reg_covar < 100 * REG_COVAR:
            return mixture_aeb(
                flat_data=flat_data,
                flat_len=flat_len_,
                means_init=means_init,
                threshold_sds=threshold_sds,
                reg_covar=reg_covar * 2,
            )
        else:
            raise exc


def well_to_aeb(well_entry=None, log: bool = True, threshold_sds=5):
    """Calculates AEB of a well based on a 2-Gaussian mixture model.

    Parameters
    ----------
    well_entry : str, iterable, or os.DirEntry, optional
        The well to be analyzed. If not provided, a filedialog will
        open to select the output file for a well.
    log : bool, default True
        Should the data be log-transformed before fitting the Gaussian
        mixture model?
    threshold_sds : numeric, default 5
        The number of standard deviations above the mean to use as a
        threshold for distinguishing on-beads in the case where there
        are very few on-beads.

    Returns
    -------
    float or dict
        AEB, based on weighted average of two measures of calculating
        on-fraction. If multiple well_entry values are provided, a
        dictionary is returned with the AEB values for each well.
    """
    if isinstance(well_entry, str):
        data = pd.read_csv(well_entry)
    elif hasattr(well_entry, "__iter__"):
        return {
            entry: well_to_aeb(entry, log=log, threshold_sds=threshold_sds)
            for entry in well_entry  # type: ignore
        }
    elif isinstance(well_entry, os.DirEntry):
        data = pd.read_csv(well_entry.path)
    else:
        well_path = filedialog.askopenfilenames(
            title="Choose a File with the Flow Cytometry Data for One Well",
            filetypes=[("Comma-Separated Values", "csv")],
        )
        return well_to_aeb(well_path)
    flat_data = log_transform(flatten(data)) if log else flatten(data)
    return mixture_aeb(flat_data=flat_data, threshold_sds=threshold_sds)


def extended_coefs(concs, aebs, corr="c4", cal_curve=None) -> dict:
    """Calculates the coefficients for a 4PL model.

    Parameters
    ----------
    concs : array-like
        The concentrations for the calibrators.
    aebs : array-like
        The AEBs for the calibrators.
    corr : {"n", "n-1", "n-1.5", "c4"} or numeric, default "c4"
        The sample standard deviation under-estimates the population
        standard deviation for a normally distributed variable.
        Specifies how this should be addressed. Options:

            - "n" : Divide by the number of samples to yield the
              uncorrected sample standard deviation.

            - "n-1" : Divide by the number of samples minus one to
              yield the square root of the unbiased sample variance.

            - "n-1.5" : Divide by the number of samples minus 1.5 to
              yield the approximate unbiased sample standard deviation.

            - "c4" : Divide by the correction factor to yield the
              exact unbiased sample standard deviation.

            - If numeric, gives the delta degrees of freedom.
    cal_curve : CalCurve, optional
        A CalCurve object to use. If not provided, a new CalCurve will
        be calculated based on the concs and aebs provided.

    Returns
    -------
    dict
        A dictionary of coefficients and properties of the calibration
        curve. Its elements are:

            "a", "b", "c", "d" : coefficients of the 4PL fit

            "LOD" : limit of detection

            "LLOQ" : lower limit of quantification (10 standard
            deviations above background)

            "blank mean" : mean AEB at 0 concentration

            "blank std" : standard deviation of AEB at 0 concentration

            "blank cv" : coefficient of variation of AEB at 0
            concentration

    """
    blank_array = flatten(aebs[concs == 0])
    if isinstance(cal_curve, CalCurve):
        new_cal_curve = cal_curve
        calibratible = True
    else:
        new_cal_curve = CalCurve()
        try:
            new_cal_curve.fit(concs, aebs)
        except Exception:
            calibratible = False
        else:
            calibratible = True
    if calibratible:
        coefs = dict(zip("ABCDEFG", new_cal_curve.coef_.items()))
        coefs["LOD"] = new_cal_curve.lod_
        coefs["LLOQ"] = limit_of_detection(
            blank_array, new_cal_curve, sds=10, corr=corr
        )
    else:
        coefs = {
            "a": np.nan,
            "b": np.nan,
            "c": np.nan,
            "d": np.nan,
            "LOD": np.inf,
            "LLOQ": np.inf,
        }
    coefs["blank mean"] = np.mean(blank_array)
    try:
        ddof = _DDOFS[corr]
    except KeyError:
        ddof = float(corr)
    corr_factor = c4(len(blank_array)) if (corr == "c4") else 1
    coefs["blank std"] = np.std(blank_array, ddof=ddof) / corr_factor
    coefs["blank cv"] = coefs["blank std"] / coefs["blank mean"]
    return coefs


def plate_subsets(
    dir_path=None,
    save_aebs_to=None,
    save_coefs_to=None,
    log: bool = True,
    model="4PL",
    lod_sds=3,
    subsets: int = 10,
    sizes=(),
    corr="c4",
    threshold_sds=5,
) -> pd.DataFrame:
    """Calculates AEBs and coefficients for a plate.

    Parameters
    ----------
    dir_path : str, optional
        The directory containing the MOSAIC plate files. If not provided
        a file dialog will be opened to select the folder.
    save_aebs_to : str, optional
        The path to save the AEBs to.
    save_coefs_to : str, optional
        The path to save the coefficients to.
    log : bool, default True
        Should the data be log-transformed before fitting the Gaussian
        mixture model?
    model : str or Model, default "4PL"
        The model to fit.
    lod_sds : numeric, default 3
        The number of standard deviations above the mean to use as a
        limit of detection.
    subsets : int, default 10
        The number of subsets to create.
    sizes : iterable
        The number of beads in each subset. Technically optional, but
        the default argument is "()" which does not conduct any
        subsetting.
    corr : {"n", "n-1", "n-1.5", "c4"} or numeric, default "c4"
        The sample standard deviation under-estimates the population
        standard deviation for a normally distributed variable.
        Specifies how this should be addressed. Options:

            - "n" : Divide by the number of samples to yield the
              uncorrected sample standard deviation.

            - "n-1" : Divide by the number of samples minus one to
              yield the square root of the unbiased sample variance.

            - "n-1.5" : Divide by the number of samples minus 1.5 to
              yield the approximate unbiased sample standard deviation.

            - "c4" : Divide by the correction factor to yield the
              exact unbiased sample standard deviation.

            - If numeric, gives the delta degrees of freedom.
    Returns
    -------
    pd.DataFrame
        A dataframe of coefficients and extended parameters for each
        subset.

    See Also
    --------
    extended_coefs : for the coefficients and parameters returned

    """
    plate_files = PlateFileCollection(dir_path)
    concs = []
    datasets = []
    dlens = []
    aebs = []
    for well in plate_files.wells:
        concs.append(plate_files.conc_map[well])
        data = pd.read_csv(plate_files.file_map[well])
        flat_data = log_transform(flatten(data.values)) if log else flatten(data.values)
        dlens.append(len(flat_data))
        datasets.append(flat_data)
        aebs.append(mixture_aeb(flat_data=flat_data, threshold_sds=threshold_sds))
    concs_flat = flatten(concs)
    aebs_flat = flatten(aebs)
    cal_curve = CalCurve(model=model, lod_sds=lod_sds).fit(concs_flat, aebs_flat)
    coefs = extended_coefs(concs_flat, aebs_flat, corr, cal_curve)
    coef_table = pd.DataFrame.from_dict(coefs, orient="index")
    aeb_table = pd.DataFrame.from_dict(
        {
            "Well": plate_files.wells,
            "Concentration": concs_flat,
            "AEB": aebs_flat,
            "Number of Beads": dlens,
        }
    )
    rng = default_rng()
    subset_aebs = {}
    for size in sizes:
        subset_aebs[size] = pd.DataFrame(
            columns=range(subsets), index=plate_files.wells, dtype=float
        )
        for w in range(len(concs)):
            if dlens[w] > size:
                for subset in range(subsets):
                    subset_data = rng.choice(datasets[w], size, replace=False)
                    subset_aebs[size].loc[plate_files.wells[w], subset] = mixture_aeb(
                        flat_data=subset_data, threshold_sds=threshold_sds
                    )
            else:
                for subset in range(subsets):
                    subset_aebs[size].loc[plate_files.wells[w], subset] = aebs_flat[w]
        subset_aebs[size]["mean"] = subset_aebs[size][range(subsets)].mean(
            axis=1, skipna=False
        )
        subset_aebs[size]["std"] = subset_aebs[size][range(subsets)].std(
            axis=1, skipna=False, ddof=1.5
        )
        subset_aebs[size]["median"] = subset_aebs[size][range(subsets)].median(
            axis=1, skipna=False
        )
        subset_aebs[size]["IQR 25%"] = subset_aebs[size][range(subsets)].quantile(
            q=0.25, axis=1
        )
        subset_aebs[size]["IQR 75%"] = subset_aebs[size][range(subsets)].quantile(
            q=0.75, axis=1
        )
    if save_aebs_to is None:
        save_aebs_to_ = filedialog.asksaveasfilename(
            initialfile=plate_files.name + " subset AEBs.xlsx",
            defaultextension=".xlsx",
            filetypes=[("Excel Workbook", "xlsx")],
        )
    else:
        save_aebs_to_ = save_aebs_to
    if save_aebs_to_:
        with pd.ExcelWriter(save_aebs_to_) as writer:
            aeb_table.to_excel(writer, sheet_name="Original", index=False)
            for size in sizes:
                subset_aebs[size].to_excel(writer, sheet_name=str(size))
    subset_coefs = {}
    for size in sizes:
        subset_coefs[size] = pd.DataFrame(
            columns=range(subsets), index=coefs_index, dtype=float
        )
        for subset in range(subsets):
            one_subset_coefs = extended_coefs(
                concs_flat, subset_aebs[size][subset], corr
            )
            for key, value in one_subset_coefs.items():
                subset_coefs[size].loc[key, subset] = value
        subset_coefs[size]["mean"] = (
            subset_coefs[size][range(subsets)].fillna(np.inf).mean(axis=1, skipna=False)
        )
        subset_coefs[size]["std"] = (
            subset_coefs[size][range(subsets)]
            .fillna(np.inf)
            .std(axis=1, skipna=False, ddof=1.5)
        )
        subset_coefs[size]["median"] = (
            subset_coefs[size][range(subsets)]
            .fillna(np.inf)
            .median(axis=1, skipna=False)
        )
        subset_coefs[size]["IQR 25%"] = (
            subset_coefs[size][range(subsets)].fillna(np.inf).quantile(q=0.25, axis=1)
        )
        subset_coefs[size]["IQR 75%"] = (
            subset_coefs[size][range(subsets)].fillna(np.inf).quantile(q=0.75, axis=1)
        )
    if save_coefs_to is None:
        save_coefs_to_ = filedialog.asksaveasfilename(
            initialfile=plate_files.name + " subset coefficients.xlsx",
            defaultextension=".xlsx",
            filetypes=[("Excel Workbook", "xlsx")],
        )
    else:
        save_coefs_to_ = save_coefs_to
    if save_coefs_to_:
        with pd.ExcelWriter(save_coefs_to_) as writer:
            coef_table.to_excel(writer, sheet_name="Original", header=False)
            for size in sizes:
                subset_coefs[size].to_excel(writer, sheet_name=str(size))
    return coef_table



import os
import pathlib
from collections.abc import Callable, Hashable, Iterable
from io import IOBase
from mmap import mmap
from tkinter import Tk, filedialog
from typing import Any, Optional

import pandas as pd

from .cal_curve import CalCurve, _CalCurveSeries
from .core import match_kwargs
from .model import MODELS, Model


def _read_tsv(*args, **kwargs):
    kwargs.setdefault("delimiter", "\t")
    return pd.read_csv(*args, **kwargs)


_EXTENSION_READERS: dict[str, Callable] = {
    ".csv": pd.read_csv,
    ".tsv": _read_tsv,
    ".xls": pd.read_excel,
    ".xlsb": pd.read_excel,
    ".xlsm": pd.read_excel,
    ".xlsx": pd.read_excel,
    "": pd.read_csv,
}

_SAMPLE_RESULTS_COLNAMES: dict[str, str] = {
    "AEB": "Replicate AEB",
    "Concentration": "Replicate Conc.",
    "Mean Concentration": "Mean Conc.",
    "SD Concentration": "SD Conc.",
    "CV Concentration": "CV Conc.",
}

_HDX_EXCEPTIONS: dict[str, Any] = {
    "TooMuchFluorescenceInResorufinChannelException": 30,
}


def _suffix(filepath_or_buffer: os.PathLike | str | IOBase) -> str:
    """Determines the file extension for a given filepath or buffer.

    Parameters
    ----------
    filepath_or_buffer : str, path object, or file-like object
        The filepath or buffer to determine the file extension for.

    Returns
    -------
    str
        The file extension, including the leading dot (e.g., '.txt').

    Raises
    ------
    TypeError
        If the input is not a valid filepath or buffer.
    """
    if hasattr(filepath_or_buffer, "suffix"):
        return filepath_or_buffer.suffix  # type: ignore
    elif isinstance(filepath_or_buffer, str):
        i = filepath_or_buffer.rfind(".")
        return filepath_or_buffer[i:] if (0 < i < len(filepath_or_buffer) - 1) else ""
    elif isinstance(filepath_or_buffer, IOBase):
        return ""
    else:
        raise TypeError(
            "Could not determine file extension for file "
            f"{filepath_or_buffer} of type {type(filepath_or_buffer)}."
        )


def _select_files_or_folders(folder: bool = False, multiple: bool = True) -> list[str]:
    """Displays a dialog box to select files or a folder.

    Parameters
    ----------
    folder : bool, default False
        If False (default), the dialog box will allow selecting files.
        If True, the dialog box will allow selecting a folder.
    multiple : bool, default True
        If True (default), the dialog box will allow selecting multiple
        files. If False, the dialog box will only allow selecting a
        single file. Ignored if folder is True.

    Returns
    -------
    io
        A list of strings representing the paths of the selected files
        or folder.

    Notes
    -----
    This function uses the `Tk` and `filedialog` modules from the
    `tkinter` package to display the dialog box. A hidden `Tk` window is
    created to facilitate the selection, and it is destroyed after the
    selection is made. The dialog box may be hidden behind other
    windows.

    """
    root = Tk()
    root.withdraw()

    if folder:
        io = [filedialog.askdirectory()]
    elif multiple:
        io = list(filedialog.askopenfilenames())
    else:
        io = [filedialog.askopenfilename()]

    root.destroy()
    return io


def _accumulate_list(iterable: Iterable, func: Callable) -> list:
    """Applies a function to iterable items and accumulates a list.

    Applies a given function to each item in an iterable and
    accumulates the results in a single list. The function should
    return an iterable for each item in the input iterable. The
    function is applied to each item, and the returned iterables are
    combined into a single list.

    Parameters
    ----------
    iterable : Iterable
        An iterable containing items that the function `func` will be
        applied to.
    func : Callable
        A function that takes an item from the iterable as input and
        returns an iterable.

    Returns
    -------
    list
        A list containing the combined results of applying the function
        `func` to each item in the input iterable.
    """
    result = []
    for item in iterable:
        result.extend(func(item))
    return result


def _is_leaf(io) -> bool:
    """Check if an object is a single file-like object."""
    return isinstance(io, (str, bytes, mmap, os.PathLike)) or not isinstance(
        io, Iterable
    )


def crawl(io) -> list:
    """Traverses directories and iterables to assemble a list of files.

    If a filepath (e.g., a string or an os.PathLike) or buffer is
    passed, it will be returned as a list. If a directory path is
    passed, then a list of all of its constituent files will be
    returned. Finally, if a collection of filepaths, buffers, or
    dictionaries is passed, then a list of all of their constituent
    files will be returned.

    Parameters
    ----------
    io : Any
        Any valid string path is acceptable. The string could be a URL.
        Valid URL schemes include http, ftp, s3, gs, and file. For file
        URLs, a host is expected. A local file could be:
        file://localhost/path/to/table.csv. If you want to pass in a
        path object, crawl accepts any os.PathLike. By file-like
        object, we refer to objects with a read() method, such as a
        file handle (e.g. via builtin open function) or StringIO. If
        the path of a directory is passed, it will be traversed and a
        list of its constituent files will be returned. If a folder or
        collection of files is passed, then each element of the
        collection will be crawled.

    Returns
    -------
    list
        A list of files or buffers. If `io` is a directory, then this
        list contains all files inside of it and its subfolders. If
        `io` is a collection of files, buffers, and directories, then
        this list contains all files and buffers in the collection, and
        all files inside of directories in the collection.
    """

    if hasattr(io, "read") or hasattr(io, "write"):
        return [io]

    elif _is_leaf(io):
        try:
            io_path = pathlib.Path(io)
        except TypeError:
            return [io]
        if io_path.is_dir():
            return _accumulate_list(io_path.iterdir(), crawl)
        return [io_path]

    else:
        return _accumulate_list(io, crawl)


def _get_sample_barcode(row: pd.Series) -> Hashable:
    """Gets the sample ID from a given row of a pandas DataFrame.

    If the "Sample Barcode" field is missing, the sample ID is
    constructed using a tuple of the "Batch Name" and "Location"
    fields. Otherwise, the "Sample Barcode" value is used as the
    sample ID.

    Parameters
    ----------
    row : pd.Series
        A row from self.raw.

    Returns
    -------
    str or tuple of str
        The sample ID.
    """
    if (
        ("Sample Barcode" in row)
        and pd.notna("Sample Barcode")
        and (row["Sample Barcode"] != "")
    ):
        return row["Sample Barcode"]
    else:
        return (row["Batch Name"], row["Location"])


class HDX:
    """Quanterix HD-X file reader.

    Reads in data from one or more Quanterix HD-X run histories
    (preferred) or sample results reports. The data are combined into a
    pandas.DataFrame called `raw`, and the assays are identified. From
    there, calibration curves can be fit to the data, and the data can
    be tidied.

    Parameters
    ----------
    io : Any, optional
        Any valid string path is acceptable. The string could be a URL.
        Valid URL schemes include http, ftp, s3, gs, and file. For file
        URLs, a host is expected. A local file could be:
        file://localhost/path/to/table.csv. If you want to pass in a
        path object, crawl accepts any os.PathLike. By file-like
        object, we refer to objects with a read() method, such as a
        file handle (e.g. via builtin open function) or StringIO. If
        the path of a directory is passed, it will be traversed and a
        list of its constituent files will be returned. If a list,
        tuple, set, dictionary, or generator is passed, then each
        element of the collection will be crawled. If not provided, a
        dialogue box will open, asking the user to select files.

    Examples
    --------
    To read in and combine Quanterix run histories and sample results
    reports chosen from a dialogue box, call without any arguments:

    >>> q = wlt.HDX()

    """

    def __init__(
        self,
        io=None,
        raw: Optional[pd.DataFrame | Iterable[pd.DataFrame | pd.Series]] = None,
        cal_curves: Optional[pd.Series] = None,
        **kwargs,
    ):
        if raw is None:
            # initialize attributes
            file_extensions = kwargs.get("file_extensions", [".csv", ".xls"])

            # read in data files
            if io is None:
                io = _select_files_or_folders(
                    folder=kwargs.get("folder", False),
                    multiple=kwargs.get("multiple", True),
                )
            filepaths = crawl(io)

            # determine which kwargs from __init__ to pass to each reader
            readers = {
                key: value
                for key, value in _EXTENSION_READERS.items()
                if key in file_extensions
            }
            reader_kwargs = {
                suffix: match_kwargs(reader, kwargs)
                for suffix, reader in readers.items()
            }

            # read in a DataFrame for each file
            raw_dfs = []
            for filepath in filepaths:
                suffix = _suffix(filepath)
                if suffix in readers:
                    raw_dfs.append(
                        readers[suffix](filepath, **reader_kwargs[suffix]).rename(
                            columns=_SAMPLE_RESULTS_COLNAMES
                        )
                    )

            # concatenate the DataFrames
            self.raw = pd.concat(raw_dfs, ignore_index=True).drop_duplicates()
        elif isinstance(raw, pd.DataFrame):
            self.raw = raw
        elif isinstance(raw, Iterable):
            self.raw = pd.concat(raw, ignore_index=True).drop_duplicates()
        else:
            raise ValueError(
                "Parameter `raw` must be a DataFrame or an iterable of DataFrames if provided."
            )
        self.raw["Sample Barcode"] = self.raw.apply(_get_sample_barcode, axis=1)  # type: ignore

        self._cal_curves = cal_curves
        self._tidy = None

        # Find and enumerate assays
        self.data_cols = kwargs.get("data_cols", ["Replicate AEB"])

        self.assay_defining_cols = kwargs.get("assay_defining_cols", ["Assay", "Plex"])

        if "assays" in kwargs:
            self.assays = kwargs["assays"]
        else:
            assays = self.raw.value_counts(subset=self.assay_defining_cols).index
            levels = reversed(assays.names[1:])
            for name in levels:
                dropped = assays.droplevel(name)
                if (assays.get_level_values(name).value_counts().shape[0] <= 1) or (
                    (name == assays.names[-1]) and dropped.nunique() == assays.nunique()
                ):
                    assays = dropped
            self.assays = assays.unique()

    # Calibration curve methods: calculate_cal_curves, cal_curves

    def calculate_cal_curves(
        self,
        model: str | Model | dict | pd.Series = "4PL",
        X_name: str = "Replicate Conc.",
        y_name: str = "Replicate AEB",
        force: bool = False,
        include_assays: Optional[Iterable] = None,
        exclude_assays: Optional[Iterable] = None,
        **kwargs,
    ) -> pd.Series:
        # get the list of assays to generate curves for
        if isinstance(model, Model) or model in MODELS:
            if include_assays is not None:
                if isinstance(include_assays, (str, tuple)):
                    assay_models = {include_assays: model}
                else:
                    assay_models = {assay: model for assay in include_assays}
            elif exclude_assays is not None:
                if isinstance(exclude_assays, (str, tuple)):
                    assay_models = {
                        assay: model for assay in self.assays if assay != exclude_assays
                    }
                else:
                    assay_models = {
                        assay: model
                        for assay in self.assays
                        if assay not in exclude_assays
                    }
            else:
                assay_models = {assay: model for assay in self.assays}
        elif hasattr(model, "items"):
            assay_models = model
        else:
            raise ValueError("Model not found.")

        # generate the curves
        cc_init_kwargs = match_kwargs(CalCurve, kwargs)
        cc_fit_kwargs = match_kwargs(CalCurve.fit, kwargs)
        self._cal_curves = _CalCurveSeries(index=self.assays, dtype=object)

        for assay, model in assay_models.items():  # type: ignore
            if isinstance(assay, tuple):  # MultiIndex
                indexer = pd.DataFrame(
                    {
                        name: self.raw[name] == level_value
                        for name, level_value in zip(self.assays.names, assay)
                    }
                ).all(axis=1)
            else:  # 1D Index
                indexer = self.raw[self.assays.name] == assay
            indexer = indexer & (self.raw["Sample Type"] == "Calibrator")

            X = self.raw.loc[indexer, X_name]
            y = self.raw.loc[indexer, y_name]
            try:
                self._cal_curves[assay] = CalCurve(model=model, **cc_init_kwargs).fit(
                    X=X, y=y, **cc_fit_kwargs
                )
            except Exception as e:
                if force:
                    raise RuntimeError(f"Error on assay {assay}.") from e
        return self._cal_curves

    @property
    def cal_curves(self) -> pd.Series:
        if self._cal_curves is None:
            self._cal_curves = self.calculate_cal_curves()  # Use default parameters
        return self._cal_curves

    # Make concentration

    def _calculate_concentrations(
        self,
        colname: Hashable = "Replicate AEB",
        newname: Optional[str] = None,
        fix_hdx_exceptions: bool = False,
        **kwargs,
    ) -> pd.Series:
        "Create a new column in self.raw and calculate concentrations."
        if (self._cal_curves is None) or kwargs:
            self.calculate_cal_curves(**kwargs)

        if fix_hdx_exceptions:

            def replicate_aeb(row):  # type: ignore
                if pd.isna(row[colname]) and row["Errors"]:
                    for key in _HDX_EXCEPTIONS:
                        if key in row["Errors"]:
                            return _HDX_EXCEPTIONS[key]
                return row[colname]

        else:

            def replicate_aeb(row):
                return row[colname]

        if len(self.assays.names) > 1:

            def get_assay(row):  # type: ignore
                return tuple(row[self.assays.names])
        else:

            def get_assay(row):
                return row[self.assays.name]

        def apply_cal_curve(row: pd.Series) -> float:
            assay = get_assay(row)
            return self.cal_curves[assay].estimate(replicate_aeb(row))

        if newname is None:
            newname = f"Concentration Calculated from {colname}"
        self.raw[newname] = self.raw.apply(apply_cal_curve, axis=1)
        return self.raw[newname]

    # Tidy data methods: _make_tidy, tidy

    def calculate_tidy(
        self,
        stat: str | Callable = "median",
        colname: Optional[str] = None,
        use_curves: bool = False,
        **kwargs,
    ):
        if colname is None:
            colname = "Replicate AEB" if use_curves else "Replicate Conc."
        specimens = self.raw
        if use_curves:
            colname = self._calculate_concentrations(colname=colname, **kwargs).name  # type: ignore

        tidy_columns = ["Sample Barcode", *list(self.assays.names), colname]
        specimens = self.raw[self.raw["Sample Type"] == "Specimen"][tidy_columns]

        self._tidy = (
            specimens.groupby(["Sample Barcode", *self.assays.names])
            .agg(stat)
            .reset_index()
            .pivot_table(
                index="Sample Barcode", columns=self.assays.names, values=colname
            )
        )

        return self._tidy

    @property
    def tidy(self) -> pd.DataFrame:
        if self._tidy is None:
            self.calculate_tidy()
        return self._tidy  # type: ignore

    def __add__(self, other):
        if not isinstance(other, HDX):
            raise TypeError("Unsupported operand type for +")

        return HDX(raw=[self.raw, other.raw])

    def __eq__(self, other) -> bool:
        """Two HDX objects are equal if they have the same raw values."""
        if not isinstance(other, HDX):
            raise TypeError("Unsupported operand type for ==")

        if len(self.raw) != len(other.raw):
            return False

        for column in self.raw.columns:
            if column in other.raw.columns and not self.raw[column].equals(
                other.raw[column]
            ):
                return False

        return True



"""Core functionality for the waltlabtools module.

Everything in waltlabtools.core is automatically imported with
waltlabtools, so it can be accessed via, e.g.,

.. code-block:: python

   import waltlabtools as wlt  # waltlabtools main functionality

   my_data = wlt.flatten([[[[[0], 1], 2], 3], 4])  # flatten a list


-----


"""

import inspect
import warnings
from functools import wraps
from typing import Any, Callable, Iterable, Literal, Optional

from sklearn.utils.validation import indexable

from ._backend import gammaln, gmean, jit, np


def deprecate(replace_with: Optional[str] = None) -> Callable:
    """Decorator to issue a DeprecationWarning with an optional replacement function."""

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            warning_message = f"{func.__name__} is deprecated as of version 1.0."
            if replace_with:
                warning_message += f" Please use {replace_with} instead."
            warnings.warn(warning_message, DeprecationWarning)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _flatten_recursively(a) -> list:
    """Recursively flatten a primitive or nested iterable.

    Parameters
    ----------
    a : Any
        The iterable to be flattened.

    Returns
    -------
    flat_a : list
        A flattened list of the input iterable's elements.

    Examples
    --------
    >>> _flatten_recursively([1, [2, [3]], 4])
    [1, 2, 3, 4]

    >>> _flatten_recursively({'a': [1, 2], 'b': {'c': 3, 'd': 4}})
    [1, 2, 3, 4]

    >>> _flatten_recursively(1)
    [1]

    """
    flat_a = []
    iterator = a.values() if isinstance(a, dict) else a
    try:
        for datum in iterator:
            flattened_datum = _flatten_recursively(datum)
            flat_a.extend(flattened_datum)
    except TypeError:
        return [a]
    else:
        return flat_a


def flatten(a, order: Literal["C", "F", "A", "K"] = "K") -> np.ndarray:
    """Flatten almost anything into a 1-dimensional numpy array.

    In simple cases, this function is a wrapper for numpy.ravel. In more
    complex cases, it recursively flattens nested iterables.

    Parameters
    ----------
    a : any
        The object to be flattened.
    order : {'C', 'F', 'A', 'K'}, optional
        The elements of `a` are read using this index order. 'C' means
        to index the elements in row-major, C-style order,
        with the last axis index changing fastest, back to the first
        axis index changing slowest.  'F' means to index the elements
        in column-major, Fortran-style order, with the
        first index changing fastest, and the last index changing
        slowest. Note that the 'C' and 'F' options take no account of
        the memory layout of the underlying array, and only refer to
        the order of axis indexing.  'A' means to read the elements in
        Fortran-like index order if `a` is Fortran *contiguous* in
        memory, C-like order otherwise.  'K' means to read the
        elements in the order they occur in memory, except for
        reversing the data when strides are negative.  By default, 'C'
        index order is used.

    Returns
    -------
    numpy.ndarray
        A 1-dimensional numpy array containing the flattened elements.

    See Also
    --------
    numpy.ravel : Flatten a numpy array.

    Notes
    -----
    When jax has been loaded as a backend, this function will raise an
    error if `a` has non-numerical elements.

    Examples
    --------
    >>> a = [[1, 2, 3], [4, 5, 6]]
    >>> flatten(a)
    array([1, 2, 3, 4, 5, 6])

    >>> b = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    >>> flatten(b)
    array([1, 2, 3, 4, 5, 6, 7, 8])

    """
    try:
        y = np.ravel(np.asarray(a, order=order), order=order)
        if y.dtype.kind == "O":
            raise ValueError("Cannot flatten object arrays.")
        return y
    except (TypeError, ValueError):
        return np.ravel(np.asarray(_flatten_recursively(a), order=order), order=order)


def coerce_array(func: Callable) -> Callable:
    """Coerce the argument to an array upon TypeError.

    This decorator is intended to wrap functions that are primarily
    called with a first argument that is compatible with `numpy.array`.
    If calling the function results in a `TypeError`, it tries coercing
    the first argument to a numpy array and calling the function again.
    This is particularly useful for functions that might be passed
    lists or other array-like objects but are designed to work with
    numpy arrays.

    Parameters
    ----------
    func : callable
        The function to be decorated.

    Returns
    -------
    callable
        The wrapped function which coerces its first argument to a
        numpy array upon TypeError.

    Examples
    --------
    >>> @coerce_array
    ... def add_one(arr):
    ...     return arr + 1

    >>> add_one([1, 2, 3])
    array([2, 3, 4])

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TypeError:
            if args:  # if the first argument is positional
                args = (np.asarray(args[0]),) + args[1:]
            else:  # if the first argument is keyword
                sig = inspect.signature(func)
                first_arg_name = next(iter(sig.parameters.keys()))
                kwargs[first_arg_name] = np.asarray(kwargs[first_arg_name])
            return func(*args, **kwargs)

    return wrapper


@coerce_array
def aeb(fon_):
    """Compute the average number of enzymes per bead.

    Converts the fraction of on-beads (fon) to the average number of
    enzymes per bead (AEB) using Poisson statistics. The formula used
    is `aeb_` = -log(1 - `fon_`).

    Parameters
    ----------
    fon_ : numeric or array-like
        A scalar or array of fractions of beads which are "on."

    Returns
    -------
    aeb_ : same as input, or array
        The average number of enzymes per bead.

    See Also
    --------
    fon : inverse of ``aeb``

    """
    return -np.log(1 - fon_)


@coerce_array
def fon(aeb_):
    """Compute the fraction of beads which are on.

    Converts the average enzymes per bead (AEB) to the fraction of
    on-beads (fon) using Poisson statistics. The formula used is
    `fon_` = 1 - exp(-`aeb_`).

    Parameters
    ----------
    aeb_ : numeric or array-like
        A scalar or array of the average number of enzymes per bead.

    Returns
    -------
    fon_ : same as input, or array
        The fractions of beads which are "on."

    See Also
    --------
    aeb : inverse of ``fon``

    """
    return 1 - np.exp(-aeb_)


@coerce_array
def c4(n):
    """Factor `c4` for unbiased estimation of normal standard deviation.

    For a finite sample, the sample standard deviation tends to
    underestimate the population standard deviation. See, e.g.,
    https://www.spcpress.com/pdf/DJW353.pdf for details. Dividing the
    sample standard deviation by the correction factor `c4` gives an
    unbiased estimator of the population standard deviation. This
    correction factor should be applied on top of Bessel's correction,
    so n-1 is used as the degrees of freedom.

    Parameters
    ----------
    n : int or array
        The number of samples.

    Returns
    -------
    numeric or array
        The correction factor, usually written `c4` or `b(n)`.

    See Also
    --------
    std : unbiased standard deviation

    numpy.std : standard deviation

    lod : limit of detection

    """
    return np.sqrt(2 / (n - 1)) * np.exp(gammaln(n / 2) - gammaln((n - 1) / 2))


_DDOFS = {"n": 0, "n-1": 1, "n-1.5": 1.5, "c4": 1}


@coerce_array
def std(
    a, corr: str | int = "c4", axis: Optional[int | tuple[int, ...]] = None, **kwargs
):
    """Compute (an unbiased estimate of) the standard deviation.

    For a finite sample, the sample standard deviation tends to
    underestimate the population standard deviation. This function
    divides the sample standard deviation by the correction factor
    `c4` to give an unbiased estimator of the population standard
    deviation.

    Parameters
    ----------
    a : array-like
        The array of values.
    corr : {"n", "n-1", "n-1.5", "c4"} or numeric, default "c4"
        The sample standard deviation under-estimates the population
        standard deviation for a normally distributed variable.
        Specifies how this should be addressed. Options:

            - "n" : Divide by the number of samples to yield the
              uncorrected sample standard deviation.

            - "n-1" : Divide by the number of samples minus one to
              yield the square root of the unbiased sample variance.

            - "n-1.5" : Divide by the number of samples minus 1.5 to
              yield the approximate unbiased sample standard deviation.

            - "c4" : Divide by the correction factor to yield the
              exact unbiased sample standard deviation.

            - If numeric, gives the delta degrees of freedom.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the standard deviation is computed. The
        default is to compute the standard deviation of the flattened array.
        If this is a tuple of ints, a standard deviation is performed over
        multiple axes, instead of a single axis or all the axes as before.

    Returns
    -------
    numeric
        The unbiased standard deviation.

    See Also
    --------
    numpy.std : standard deviation

    c4 : correction factor used when corr="c4"

    """
    ddof = _DDOFS.get(corr, corr)  # type: ignore
    ret = np.std(a, axis=axis, ddof=ddof, **kwargs)  # type: ignore

    if corr == "c4":
        n = np.prod(np.shape(a))
        if isinstance(axis, int):
            n /= np.prod(np.shape(a)[axis])
        elif isinstance(axis, Iterable):
            n /= np.prod([np.shape(a)[ax] for ax in axis])
        return ret / c4(n)
    else:
        return ret


@jit
def _gmnd_f_step(
    previous_values: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08
) -> tuple[np.ndarray, bool]:
    """Compute one step of the Geothmetic Meandian algorithm.

    Parameters
    ----------
    previous_values : numpy.ndarray
        An array containing the previous values of the Geothmetic
        Meandian algorithm.
    rtol : float, default 1e-05
        The relative tolerance parameter used to determine convergence.
    atol : float, default 1e-08
        The absolute tolerance parameter used to determine convergence.

    Returns
    -------
    next_values : numpy.ndarray
        The next values of the Geothmetic Meandian algorithm.

    converged : bool
        Whether the algorithm has converged.

    Notes
    -----
    This function is decomposed this way in order to make use of
    just-in-time (JIT) compilation with Numba or Jax.
    """
    next_values = np.asarray(
        (
            np.mean(previous_values),
            gmean(previous_values),
            np.median(previous_values),
        )
    )
    converged = np.allclose(
        next_values,
        np.full_like(next_values, fill_value=next_values[2]),
        rtol=rtol,
        atol=atol,
    )
    return next_values, converged


@coerce_array
def geothmetic_meandian(
    a,
    weights=None,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    nan_policy="omit",
) -> float:
    """Compute he Geothmetic Meandian of the input data, as per XKCD.

    For details, see https://xkcd.com/2435/. This function compares
    the three most common measures of central tendency: the arithmetic
    mean, the geometric mean, and the median.
    The geometric meandian uses an iterative process that stops when
    the arithmetic mean, geometric mean, and median converge within
    (`atol` + `rtol` * their magnitudes).

    Parameters
    ----------
    data : array_like
        The input data, which can be any array-like object.
    rtol : float, default 1e-05
        The relative tolerance parameter used to determine convergence.
    atol : float, default 1e-08
        The absolute tolerance parameter used to determine convergence.

    Returns
    -------
    float
        The Geothmetic Meandian of the input data.

    Examples
    --------
    >>> geothmetic_meandian([1, 1, 2, 3, 5])
    2.089440951883
    """
    if nan_policy == "omit":
        a = dropna(a)

    current_values = np.asarray(
        (np.average(a, weights=weights), gmean(a, weights=weights), np.median(a))
    )

    converged = False
    while not converged:
        current_values, converged = _gmnd_f_step(current_values, rtol=rtol, atol=atol)
    return float(current_values[2])


def match_kwargs(
    func: Callable | Iterable[Callable] | str,
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    """Match keyword arguments to the parameters of a given function.

    Given a function and a dictionary of keyword arguments, return a
    dictionary of only those keyword arguments that match the parameters
    of the function. Typical usage is

    Parameters
    ----------
    func : callable or iterable of callable
        The function or functions to match keyword arguments against.
        If multiple functions, keywords will be matched against any
        of them (i.e., the union of all the functions' parameter names).
    kwargs : dict[str, Any]
        The dictionary of keyword arguments to match.

    Returns
    -------
    dict
        A dictionary of keyword arguments that match the parameters of
        the function, or the parameters of any of the functions.

    Notes
    -----
    The implementation of this function is based on inspect.signature.
    Some functions accept a generic `**kwargs` parameter, which is
    not included in the signature. This function does not handle such
    cases.

    Examples
    --------
    To call a function `func` on only the `kwargs` it accepts:

    >>> func(..., **_match_kwargs(func, kwargs))

    """
    if isinstance(func, str):
        return {
            key.removeprefix(func): value
            for key, value in kwargs.items()
            if key.startswith(func)
        }
    elif callable(func):
        params = [inspect.signature(func).parameters.keys()]
    else:
        params = [inspect.signature(f).parameters.keys() for f in func]
    intersected_kw_names = set(kwargs.keys()).intersection(set().union(*params))
    return {kw_name: kwargs[kw_name] for kw_name in intersected_kw_names}


def _dims_occurring_once(shape: tuple[int, ...]) -> set[int]:
    """Get dimensions that occur only once in the provided shape.

    Parameters
    ----------
    shape : Tuple[int]
        A tuple representing the shape of a numpy array.

    Returns
    -------
    set[int]
        A set of dimensions that occur only once in the input shape.
    """
    unique, counts = np.unique(np.asarray(shape), return_counts=True)
    return set(unique[counts == 1].tolist())


def _find_common_dimension(*args: np.ndarray, common_len=None) -> tuple[int, list[int]]:
    """Find the unique, common dimension length in all given arrays.

    This function identifies dimensions that appear only once in each shape
    and are common to all provided arrays.

    Parameters
    ----------
    *args : np.ndarray
        Variable length argument list of numpy arrays.

    Returns
    -------
    Tuple[int, List[int]]
        A tuple containing the common dimension length that occurs only
        once in all the arrays and a list of indices where this common
        dimension length occurs in each array's shape.

    Raises
    ------
    ValueError
        If no unique non-zero, non-unitary common dimensions are found
        between the arrays.
    """
    shapes = [arg.shape for arg in args]

    if common_len is None:
        individual_unique_dims = [_dims_occurring_once(shp) for shp in shapes]
        # Find common dimensions
        common_len = set.intersection(*individual_unique_dims) - {0, 1}
        if len(common_len) != 1:
            raise ValueError(
                "No unique non-zero, non-unitary common dimensions found "
                f"between arrays of shapes {shapes}."
            )

        common_len = common_len.pop()

    indices = [shape.index(common_len) for shape in shapes]
    return common_len, indices


def dropna(*args: Any, drop_inf: bool = False, common_len=None) -> tuple:
    """Drop rows containing NA values in any of the provided arrays.

    This function is designed for multi-dimensional numpy arrays (and objects convertible
    to them, like pandas Series). It assumes that the arrays share a common dimension
    (like rows in a 2D array) and drops any "row" with NaN or Inf values across the arrays.

    Parameters
    ----------
    *args : array-like
        Variable length argument list of objects. Expected to be numpy arrays
        or objects convertible to numpy arrays (e.g., pandas Series).
    drop_inf : bool, default False
        Whether to drop Inf values in addition to NaN values.

    Returns
    -------
    Tuple[Union[np.ndarray, pd.Series]]
        Tuple of arrays with the same dimensions as the input arrays, but with
        rows containing NaN or Inf values dropped. If the input was a pandas
        Series, the return type for that specific input will also be a pandas Series.

    Examples
    --------
    Consider two numpy arrays of shape (2, 3) and (2,):
    >>> a = np.array([[1, 2, np.nan], [4, 5, 6]])
    >>> b = np.array([7, np.nan])
    >>> dropna(a, b)
    (array([[4., 5., 6.]]), array([7.]))

    For arrays with a shape like (1, 14):
    The output retains the same number of dimensions, so an input of (1, 14)
    will produce an output shape like (1, n) depending on the number of NaN/Inf values.

    Note
    ----
    The function identifies a "common row dimension" which it uses to check for
    NaN or Inf values. This dimension is identified as a unique size that's present
    in all provided arrays.
    """
    arrs = [np.asarray(arg) for arg in args]

    common_len, common_dims = _find_common_dimension(*arrs, common_len=common_len)

    keep_func = np.isfinite if drop_inf else lambda x: np.logical_not(np.isnan(x))

    # Build masks for each array separately
    masks = []
    for arr, dim in zip(arrs, common_dims):
        moved_arr = np.moveaxis(arr, dim, 0)
        mask = keep_func(moved_arr).all(axis=tuple(range(1, arr.ndim)))
        masks.append(mask)

    # Combine masks to determine rows to keep
    keep_mask = np.all(masks, axis=0)

    arrs_na_dropped = (
        np.compress(keep_mask, arr, axis=dim) for arr, dim in zip(arrs, common_dims)
    )

    return tuple(arrs_na_dropped)


def dropna(
    *args,
    drop_inf: bool = False,
    axis: Optional[int | Iterable[int]] = 0,
    common_len: Optional[int] = None,
):
    keep_func = np.isfinite if drop_inf else lambda x: np.logical_not(np.isnan(x))

    if axis == 0:
        args = indexable(*args)  # type: ignore
        masks = []
        for arg in args:
            mask = np.all(keep_func(arg), axis=tuple(range(1, np.ndim(arg))))
            masks.append(mask)
        keep_mask = np.all(masks, axis=0)
        args_na_dropped = (np.compress(keep_mask, arg, axis=0) for arg in args)
    else:
        raise NotImplementedError
    return tuple(args_na_dropped)


def _get_value_or_key(d: dict, key: Any, t: Optional[type | tuple] = None) -> Any:
    """Look up a value, or return the key if it is of a matching type.

    If the key is found in the dictionary, its value is returned. If
    not, the key itself is returned if it is of type t (if specified) or
    the same type as any of the dictionary's values (if not). If the key
    is not found and not of the specified type(s), a KeyError is raised.

    Parameters
    ----------
    d : dict
        The dictionary from which to retrieve the value.
    key : Any
        The key to look up in the dictionary.
    t : type or tuple of types, optional
        The expected type or types of the key. If not provided
        (default), the type(s) will be inferred from the values in the
        dictionary.

    Returns
    -------
    Any
        The value associated with the key in the dictionary, or the key
        itself if the key is not found and is of the specified type(s).
    """
    try:
        return d[key]
    except (TypeError, KeyError) as e:
        if t is None:
            t = tuple({type(v) for v in d.values()})
        if isinstance(key, t):
            return key
        raise KeyError(f"key {key} not recognized and not of type {t}") from e



import warnings
from collections.abc import Iterable
from typing import Optional

import numpy as np
from sklearn.utils.validation import indexable


def _match_axes_size(shapes, common_len):
    """Helper function to choose axes based on the common length."""
    num_axes = [np.sum(np.array(shape) == common_len) > 1 for shape in shapes]
    if any(num_axes):
        warnings.warn(
            "dropna input includes square array; choosing the "
            f"first axis with length {common_len} from shapes {shapes}"
        )
    axis = [np.argmax(np.array(shape) == common_len) for shape in shapes]
    return axis


def _find_axes(*args) -> tuple[tuple | list, list]:
    try:
        args = indexable(*args)
    except ValueError as e:
        shapes = [np.shape(arg) for arg in args]
        common_lengths = set.intersection(*(set(shape) for shape in shapes))
        if len(common_lengths) == 0:
            raise ValueError(
                f"dropna found input variables with irreconcilable shapes: {shapes}"
            ) from e
        elif 0 in common_lengths:
            raise ValueError(
                f"dropna: all input arrays are empty: they have shapes {shapes}"
            ) from e
        elif len(common_lengths) == 1:
            common_len = common_lengths.pop()
            axis = _match_axes_size(shapes, common_len)
        else:  # len(common_lengths) > 1
            first_axis = (np.argmax(np.array(shape) > 1) for shape in shapes)
            first_lengths = {
                shape[ax] for shape, ax in zip(shapes, first_axis)
            }.intersection(common_lengths)
            common_len = first_lengths.pop()
            if first_lengths:
                warnings.warn(
                    "dropna was given arrays with common lengths "
                    f"{common_lengths}; choosing {common_len} for shapes {shapes}"
                )
            axis = _match_axes_size(shapes, common_len)
    else:
        axis = [0] * len(args)
    return args, axis


def align(
    *args,
    drop_inf: bool = False,
    axis: Optional[int | Iterable[int]] = None,
    common_len: Optional[int] = None,
):
    if common_len is None and axis is None:
        args, axis = _find_axes(*args)
    elif axis is not None and common_len is None:
        if isinstance(axis, int):
            axis = [axis] * len(args)
        common_lengths = {np.shape(arg)[ax] for arg, ax in zip(args, axis)}
        if len(common_lengths) == 1:
            common_len = common_lengths.pop()
        else:
            raise ValueError(
                "dropna found input variables with different lengths "
                f"{common_lengths} along the provided axis: {axis}"
            )
    elif axis is not None and common_len is not None:
        if isinstance(axis, int):
            axis = [axis] * len(args)
        common_lengths = {np.shape(arg)[ax] for arg, ax in zip(args, axis)}
        if (len(common_lengths) != 1) or (common_len not in common_lengths):
            raise ValueError(
                f"dropna found input variables whose lengths {common_lengths} "
                f"along the provided axis {axis} did not match common_len={common_len}"
            )
    elif common_len is not None and axis is None:
        shapes = [np.shape(arg) for arg in args]

        axis = _match_axes_size(shapes, common_len)

    return args, axis, common_len

    # find the axis along which to drop
    #     axes: Iterable[int], length = len(args)
    #     common_len: int = np.shape(arg)[ax] for arg, ax in zip(args, axes)
    # assemble the mask
    #     mask: ArrayLike[bool], shape (common_len,)
    # for each arg:
    #     drop rows


def dropna(
    *args,
    drop_inf: bool = False,
    axis: Optional[int | Iterable[int]] = 0,
    common_len: Optional[int] = None,
):
    keep_func = np.isfinite if drop_inf else lambda x: np.logical_not(np.isnan(x))

    if axis == 0:
        args = indexable(args)
        masks = []
        for arg in args:
            mask = np.all(keep_func(arg), axis=tuple(range(1, np.ndim(arg))))
            masks.append(mask)
        keep_mask = np.all(masks, axis=0)
        args_na_dropped = (np.compress(keep_mask, arg, axis=0) for arg in args)

    return tuple(args_na_dropped)



from collections.abc import Callable
from typing import Any

import numpy as np
from matplotlib.scale import ScaleBase

from .core import coerce_array


class Model:
    """Mathematical model for calibration curve fitting.

    A Model is an object with a function and its inverse, with one
    or more free parameters that can be fit to calibration curve data.

    Parameters
    ----------
    func : function
        Forward functional form, mapping levels (e.g., concentrations)
        to signal values (e.g., AEB). Should be a function which takes
        in `X` and other parameters and returns `y`. The first
        parameter of func should be `X`, and the remaining parameters
        should be the coefficients which are fit to the data (typically
        floats).
    inverse : function
        Inverse functional form, mapping signal values (e.g., AEB) to
        levels (e.g., concentrations). Should be a function which takes
        in `y` and other parameters and returns `X`. The first
        parameter of inverse should be `y`, and the remaining
        parameters should be the same coefficients as in fun.
    name : str
        The name of the function. For example, "4PL" or "linear".
    params : list-like of str
        The names of the parameters for the function. This should be
        the same length as the number of arguments which fun and
        inverse take after their inputs `x` and `y`, respectively.
    xscale, yscale : {"linear", "log", "symlog", "logit"} or
    matplotlib.ScaleBase, default "linear"
        The natural scaling transformations for `x` and `y`. For
        example, "log" means that the data may be distributed
        log-normally and are best visualized on a log scale.
    jac : {'2-point', '3-point', 'cs', callable}, optional
        Method of computing the Jacobian matrix (an m-by-n matrix, where
        element (i, j) is the partial derivative of f[i] with respect to
        x[j]) of the loss function. The keywords select a finite
        difference scheme for numerical estimation. The scheme '3-point'
        is more accurate, but requires twice as many operations as
        '2-point' (default). The scheme 'cs' uses complex steps, and
        while potentially the most accurate, it is applicable only when
        `func` correctly handles complex inputs and can be analytically
        continued to the complex plane. If callable, it is used as
        ``jac(x, *args, **kwargs)`` and should return a good approximation
        (or the exact value) for the Jacobian as an array_like (np.atleast_2d
        is applied), a sparse matrix (csr_matrix preferred for performance) or
        a `scipy.sparse.linalg.LinearOperator`.

    """

    def __init__(
        self,
        func: Callable,
        inverse: Callable,
        coef_init: Any,
        name: str = "",
        plaintext_formula: str = "",
        xscale: str | ScaleBase = "linear",
        yscale: str | ScaleBase = "linear",
        jac: str | Callable = "2-point",
    ):
        self.func = func
        self.inverse = inverse
        self.coef_init = coef_init
        self.name = name
        self.plaintext_formula = plaintext_formula
        self.xscale = xscale
        self.yscale = yscale
        self.coef_init = coef_init
        self.jac = jac


# LINEAR
# @coerce_array
def linear(X, a: float = 1, b: float = 0):
    return a * X + b


# @coerce_array
def linear_inverse(y, a: float = 1, b: float = 0):
    return (y - b) / a


def jac_linear(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    J_a = X * sample_weight
    J_b = sample_weight
    return np.column_stack((J_a, J_b))


linear_model = Model(
    func=linear,
    inverse=linear_inverse,
    coef_init=np.array([1, 0]),
    name="linear",
    plaintext_formula="y = a x + b",
    xscale="linear",
    yscale="linear",
    jac=jac_linear,
)


# POWER
@coerce_array
def power(X, a: float = 1, b: float = 1):
    return b * X**a


@coerce_array
def power_inverse(y, a: float = 1, b: float = 1):
    return (y / b) ** (1 / a)


def jac_power(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, b = coef
    J_a = b * X**a * np.where(X > 0, np.log(X), 0) * sample_weight
    J_b = X**a * sample_weight
    return np.column_stack((J_a, J_b))


power_model = Model(
    func=power,
    inverse=power_inverse,
    coef_init=np.array([1, 1]),
    name="power",
    plaintext_formula="y = a x^b",
    xscale="log",
    yscale="log",
    jac=jac_power,
)


# HILL
def hill(X, a: float = 1, b: float = 1, c: float = 1):
    return (a * X**b) / (c**b + X**b)


def hill_inverse(y, a: float = 1, b: float = 1, c: float = 1):
    return c * (a / y - 1) ** (-1 / b)


def jac_hill(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, b, c = coef
    X_b = X**b
    c_b = c**b
    denom = (c_b + X_b) ** 2

    J_a = X_b / (c_b + X_b)
    J_b = a * X_b * np.where(X > 0, np.log(X), 0) * (c_b - X_b) / denom
    J_c = -a * X_b * b * c ** (b - 1) / denom
    return np.column_stack((J_a, J_b, J_c)) * sample_weight[:, np.newaxis]


hill_model = Model(
    func=hill,
    inverse=hill_inverse,
    coef_init=np.array([1, 1, 1]),
    name="Hill",
    plaintext_formula="y = a x^b / (c^b + x^b)",
    xscale="log",
    yscale="log",
    jac=jac_hill,
)


# LOGISTIC
@coerce_array
def logistic(X, a: float = 1, b: float = 1, c: float = 0, d: float = 0):
    return d + (a - d) / (1 + np.exp(-b * (X - c)))


@coerce_array
def logistic_inverse(y, a: float = 1, b: float = 1, c: float = 0, d: float = 0):
    return c - np.log((a - d) / (y - d) - 1) / b


def jac_logistic(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, b, c, d = coef
    exp_term = np.exp(-b * (X - c))
    denom = 1 + exp_term
    denom2 = denom**2

    J_a = 1 / denom
    J_b = (a - d) * (X - c) * exp_term / denom2
    J_c = (a - d) * b * exp_term / denom2
    J_d = exp_term / denom
    return np.column_stack((J_a, J_b, J_c, J_d)) * sample_weight[:, np.newaxis]


logistic_model = Model(
    func=logistic,
    inverse=logistic_inverse,
    coef_init=np.array([1, 1, 0, 0]),
    name="logistic",
    plaintext_formula="y = d + (a - d) / {1 + exp[-b (x - c)]}",
    xscale="linear",
    yscale="linear",
    jac=jac_logistic,
)


# 3PL
@coerce_array
def three_param_logistic(X, a: float = 0, c: float = 1, d: float = 30):
    return d + (a - d) / (1 + X / c)


@coerce_array
def three_param_logistic_inverse(y, a: float = 0, c: float = 1, d: float = 30):
    return c * ((a - d) / (y - d) - 1)


def jac_three_param_logistic(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, c, d = coef
    denom = 1 + X / c

    J_a = 1 / denom
    J_c = (a - d) * X / (c**2 * denom**2)
    J_d = X / (c + X)

    return np.column_stack((J_a, J_c, J_d)) * sample_weight[:, np.newaxis]


three_param_logistic_model = Model(
    func=three_param_logistic,
    inverse=three_param_logistic_inverse,
    coef_init=np.array([0, 1, 30]),
    name="3PL",
    plaintext_formula="y = d + (a - d) / (1 + x/c)",
    xscale="log",
    yscale="log",
    jac=jac_three_param_logistic,
)


# 4PL
@coerce_array
def four_param_logistic(X, a: float = 0, b: float = 1, c: float = 1, d: float = 30):
    return d + (a - d) / (1 + (X / c) ** b)


@coerce_array
def four_param_logistic_inverse(
    y, a: float = 0, b: float = 1, c: float = 1, d: float = 30
):
    return c * ((a - d) / (y - d) - 1) ** (1 / b)


def jac_four_param_logistic(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, b, c, d = coef
    X_div_c = X / c
    X_div_c_b = X_div_c**b
    denom = 1 + X_div_c_b
    denom2 = denom**2

    J_a = 1 / denom
    J_b = -(a - d) * np.where(X > 0, X_div_c_b * np.log(X_div_c) / denom2, 0)
    J_c = (a - d) * b * X**b / (c ** (b + 1) * denom2)
    J_d = 1 - J_a
    return np.column_stack((J_a, J_b, J_c, J_d)) * sample_weight[:, np.newaxis]


four_param_logistic_model = Model(
    func=four_param_logistic,
    inverse=four_param_logistic_inverse,
    coef_init=np.array([0, 1, 1, 30]),
    name="4PL",
    plaintext_formula="y = d + (a - d) / [1 + (x/c)^b]",
    xscale="log",
    yscale="log",
    jac=jac_four_param_logistic,
)


# 5PL
@coerce_array
def five_param_logistic(
    X, a: float = 0, b: float = 1, c: float = 1, d: float = 30, g: float = 1
):
    return d + (a - d) / (1 + (X / c) ** b) ** g


@coerce_array
def five_param_logistic_inverse(
    y, a: float = 0, b: float = 1, c: float = 1, d: float = 30, g: float = 1
):
    return c * (((a - d) / (y - d)) ** (1 / g) - 1) ** (1 / b)


def jac_five_param_logistic(
    coef: np.ndarray, *, X: np.ndarray, sample_weight: np.ndarray
) -> np.ndarray:
    a, b, c, d, g = coef
    X_div_c = X / c
    X_div_c_b = X_div_c**b
    denom = (1 + X_div_c_b) ** g
    denom_g_plus_1 = (1 + X_div_c_b) ** (g + 1)

    J_a = 1 / denom
    J_b = (
        -(a - d) * g * X_div_c_b * np.where(X > 0, np.log(X_div_c), 0) / denom_g_plus_1
    )

    # Derivative with respect to c
    J_c = (a - d) * g * b * X**b / (c ** (b + 1) * denom_g_plus_1)

    # Derivative with respect to d
    J_d = 1 - J_a

    # Derivative with respect to g
    J_g = -(a - d) * np.log(1 + X_div_c_b) / denom

    # Stack the derivatives and apply sample weights
    J = np.column_stack((J_a, J_b, J_c, J_d, J_g))
    return J * sample_weight[:, np.newaxis]


five_param_logistic_model = Model(
    func=five_param_logistic,
    inverse=five_param_logistic_inverse,
    coef_init=np.array([0, 1, 1, 30, 1]),
    name="5PL",
    plaintext_formula="y = d + (a - d) / [1 + (x/c)^b]^g",
    xscale="log",
    yscale="log",
    jac=jac_five_param_logistic,
)


MODELS: dict[str, Model] = {
    "linear": linear_model,
    "power": power_model,
    "Hill": hill_model,
    "logistic": logistic_model,
    "3PL": three_param_logistic_model,
    "4PL": four_param_logistic_model,
    "5PL": five_param_logistic_model,
}
"""Built-in regression models.

Keys are strings giving model names; values are waltlabtools.Model 
objects.

Models
------
"linear" : Linear function.

"power" : Power function.

"Hill" : Hill function.

"logistic" : Logistic function.

"3PL" : Four-parameter logistic (3PL) function.

"4PL" : Four-parameter logistic (4PL) function.

"5PL" : Five-parameter logistic (5PL) function.

"""
